#include <blade.h>
#include <curl/curl.h>
#include <stdlib.h>
#include <math.h>

#define DEFINE_CURL_CONSTANT(v) \
  b_value __curl_##v(b_vm *vm) { \
    return NUMBER_VAL((double)(v)); \
  }

#define DEFINE_CURL_STR_CONSTANT(v) \
  b_value __curl_##v(b_vm *vm) { \
    return STRING_VAL(v); \
  }

#define DEFINE_CURL_PTR_CONSTANT(v, j) \
  b_value __curl_##v(b_vm *vm) { \
    return OBJ_VAL(new_ptr(vm, (void*)v j)); \
  }

#define GET_CURL_CONSTANT(v) \
  {#v, true, __curl_##v}

#ifndef CURL_MAX_INPUT_LENGTH
#define CURL_MAX_INPUT_LENGTH 8000000
#endif

void b__curl_module_preloader(b_vm *vm) {
  curl_global_init(CURL_GLOBAL_ALL);
}

void b__curl_module_unloader(b_vm *vm) {
  curl_global_cleanup();
}

static int CurlInfoTypes[58] = {
    CURLINFO_EFFECTIVE_URL,
    CURLINFO_RESPONSE_CODE,
    CURLINFO_TOTAL_TIME,
    CURLINFO_NAMELOOKUP_TIME,
    CURLINFO_CONNECT_TIME,
    CURLINFO_PRETRANSFER_TIME,
    CURLINFO_SIZE_UPLOAD,
    CURLINFO_SIZE_UPLOAD_T,
    CURLINFO_SIZE_DOWNLOAD,
    CURLINFO_SIZE_DOWNLOAD_T,
    CURLINFO_SPEED_DOWNLOAD,
    CURLINFO_SPEED_DOWNLOAD_T,
    CURLINFO_SPEED_UPLOAD,
    CURLINFO_SPEED_UPLOAD_T,
    CURLINFO_HEADER_SIZE,
    CURLINFO_REQUEST_SIZE,
    CURLINFO_SSL_VERIFYRESULT,
    CURLINFO_FILETIME,
    CURLINFO_FILETIME_T,
    CURLINFO_CONTENT_LENGTH_DOWNLOAD,
    CURLINFO_CONTENT_LENGTH_DOWNLOAD_T,
    CURLINFO_CONTENT_LENGTH_UPLOAD,
    CURLINFO_CONTENT_LENGTH_UPLOAD_T,
    CURLINFO_STARTTRANSFER_TIME,
    CURLINFO_CONTENT_TYPE,
    CURLINFO_REDIRECT_TIME,
    CURLINFO_REDIRECT_COUNT,
    CURLINFO_HTTP_CONNECTCODE,
    CURLINFO_HTTPAUTH_AVAIL,
    CURLINFO_PROXYAUTH_AVAIL,
    CURLINFO_OS_ERRNO,
    CURLINFO_NUM_CONNECTS,
    CURLINFO_COOKIELIST,
    CURLINFO_FTP_ENTRY_PATH,
    CURLINFO_REDIRECT_URL,
    CURLINFO_PRIMARY_IP,
    CURLINFO_APPCONNECT_TIME,
    CURLINFO_CERTINFO,
    CURLINFO_CONDITION_UNMET,
    CURLINFO_RTSP_SESSION_ID,
    CURLINFO_RTSP_CLIENT_CSEQ,
    CURLINFO_RTSP_SERVER_CSEQ,
    CURLINFO_RTSP_CSEQ_RECV,
    CURLINFO_PRIMARY_PORT,
    CURLINFO_LOCAL_IP,
    CURLINFO_LOCAL_PORT,
    CURLINFO_HTTP_VERSION,
    CURLINFO_PROXY_SSL_VERIFYRESULT,
    CURLINFO_PROTOCOL,
    CURLINFO_SCHEME,
    CURLINFO_TOTAL_TIME_T,
    CURLINFO_NAMELOOKUP_TIME_T,
    CURLINFO_CONNECT_TIME_T,
    CURLINFO_PRETRANSFER_TIME_T,
    CURLINFO_STARTTRANSFER_TIME_T,
    CURLINFO_REDIRECT_TIME_T,
    CURLINFO_APPCONNECT_TIME_T,
    CURLINFO_RETRY_AFTER
};

struct CURLInfoTypeMap {
  int type;
} CURLInfoTypeMap[] = {
    [CURLINFO_EFFECTIVE_URL] = { CURLINFO_STRING },
    [CURLINFO_RESPONSE_CODE] = { CURLINFO_LONG },
    [CURLINFO_TOTAL_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_NAMELOOKUP_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_CONNECT_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_PRETRANSFER_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_SIZE_UPLOAD] = { CURLINFO_DOUBLE },
    [CURLINFO_SIZE_UPLOAD_T] = { CURLINFO_OFF_T },
    [CURLINFO_SIZE_DOWNLOAD] = { CURLINFO_DOUBLE },
    [CURLINFO_SIZE_DOWNLOAD_T] = { CURLINFO_OFF_T },
    [CURLINFO_SPEED_DOWNLOAD] = { CURLINFO_DOUBLE },
    [CURLINFO_SPEED_DOWNLOAD_T] = { CURLINFO_OFF_T },
    [CURLINFO_SPEED_UPLOAD] = { CURLINFO_DOUBLE },
    [CURLINFO_SPEED_UPLOAD_T] = { CURLINFO_OFF_T },
    [CURLINFO_HEADER_SIZE] = { CURLINFO_LONG },
    [CURLINFO_REQUEST_SIZE] = { CURLINFO_LONG },
    [CURLINFO_SSL_VERIFYRESULT] = { CURLINFO_LONG },
    [CURLINFO_FILETIME] = { CURLINFO_LONG },
    [CURLINFO_FILETIME_T] = { CURLINFO_OFF_T },
    [CURLINFO_CONTENT_LENGTH_DOWNLOAD] = { CURLINFO_DOUBLE },
    [CURLINFO_CONTENT_LENGTH_DOWNLOAD_T] = { CURLINFO_OFF_T },
    [CURLINFO_CONTENT_LENGTH_UPLOAD] = { CURLINFO_DOUBLE },
    [CURLINFO_CONTENT_LENGTH_UPLOAD_T] = { CURLINFO_OFF_T },
    [CURLINFO_STARTTRANSFER_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_CONTENT_TYPE] = { CURLINFO_STRING },
    [CURLINFO_REDIRECT_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_REDIRECT_COUNT] = { CURLINFO_LONG },
    [CURLINFO_HTTP_CONNECTCODE] = { CURLINFO_LONG },
    [CURLINFO_HTTPAUTH_AVAIL] = { CURLINFO_LONG },
    [CURLINFO_PROXYAUTH_AVAIL] = { CURLINFO_LONG },
    [CURLINFO_OS_ERRNO] = { CURLINFO_LONG },
    [CURLINFO_NUM_CONNECTS] = { CURLINFO_LONG },
    [CURLINFO_COOKIELIST] = { CURLINFO_SLIST },
    [CURLINFO_FTP_ENTRY_PATH] = { CURLINFO_STRING },
    [CURLINFO_REDIRECT_URL] = { CURLINFO_STRING },
    [CURLINFO_PRIMARY_IP] = { CURLINFO_STRING },
    [CURLINFO_APPCONNECT_TIME] = { CURLINFO_DOUBLE },
    [CURLINFO_CERTINFO] = { CURLINFO_PTR },
    [CURLINFO_CONDITION_UNMET] = { CURLINFO_LONG },
    [CURLINFO_RTSP_SESSION_ID] = { CURLINFO_STRING },
    [CURLINFO_RTSP_CLIENT_CSEQ] = { CURLINFO_LONG },
    [CURLINFO_RTSP_SERVER_CSEQ] = { CURLINFO_LONG },
    [CURLINFO_RTSP_CSEQ_RECV] = { CURLINFO_LONG },
    [CURLINFO_PRIMARY_PORT] = { CURLINFO_LONG },
    [CURLINFO_LOCAL_IP] = { CURLINFO_STRING },
    [CURLINFO_LOCAL_PORT] = { CURLINFO_LONG },
    [CURLINFO_HTTP_VERSION] = { CURLINFO_LONG },
    [CURLINFO_PROXY_SSL_VERIFYRESULT] = { CURLINFO_LONG },
    [CURLINFO_PROTOCOL] = { CURLINFO_LONG },
    [CURLINFO_SCHEME] = { CURLINFO_STRING },
    [CURLINFO_TOTAL_TIME_T] = { CURLINFO_OFF_T },
    [CURLINFO_NAMELOOKUP_TIME_T] = { CURLINFO_OFF_T  },
    [CURLINFO_CONNECT_TIME_T] = { CURLINFO_OFF_T  },
    [CURLINFO_PRETRANSFER_TIME_T] = { CURLINFO_OFF_T  },
    [CURLINFO_STARTTRANSFER_TIME_T] = { CURLINFO_OFF_T  },
    [CURLINFO_REDIRECT_TIME_T] = { CURLINFO_OFF_T  },
    [CURLINFO_APPCONNECT_TIME_T] = { CURLINFO_OFF_T  },
    [CURLINFO_RETRY_AFTER] = { CURLINFO_OFF_T  }
};

static const char* CURLProtoMap[] = {
    "http", "https", "ftp", "ftps", "scp", "sftp", "telnet", "ldap", "ldaps",
    "dict", "file", "tftp", "imap", "imaps", "pop3", "pop3s", "smtp", "smtps",
    "rtsp", "rtmp", "rtmpt", "rtmpe", "rtmpte", "rtmps", "rtmpts", "gopher",
    "smb", "smbs", "mqtt", "gophers",
};


// CURL OPTS
DEFINE_CURL_CONSTANT(CURLOPT_URL)
DEFINE_CURL_CONSTANT(CURLOPT_PORT)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY)
DEFINE_CURL_CONSTANT(CURLOPT_USERPWD)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYUSERPWD)
DEFINE_CURL_CONSTANT(CURLOPT_RANGE)
DEFINE_CURL_CONSTANT(CURLOPT_ERRORBUFFER)
DEFINE_CURL_CONSTANT(CURLOPT_TIMEOUT)
DEFINE_CURL_CONSTANT(CURLOPT_INFILESIZE)
DEFINE_CURL_CONSTANT(CURLOPT_POSTFIELDS)
DEFINE_CURL_CONSTANT(CURLOPT_REFERER)
DEFINE_CURL_CONSTANT(CURLOPT_FTPPORT)
DEFINE_CURL_CONSTANT(CURLOPT_USERAGENT)
DEFINE_CURL_CONSTANT(CURLOPT_LOW_SPEED_LIMIT)
DEFINE_CURL_CONSTANT(CURLOPT_LOW_SPEED_TIME)
DEFINE_CURL_CONSTANT(CURLOPT_RESUME_FROM)
DEFINE_CURL_CONSTANT(CURLOPT_COOKIE)
DEFINE_CURL_CONSTANT(CURLOPT_HTTPHEADER)
DEFINE_CURL_CONSTANT(CURLOPT_HTTPPOST)
DEFINE_CURL_CONSTANT(CURLOPT_SSLCERT)
DEFINE_CURL_CONSTANT(CURLOPT_KEYPASSWD)
DEFINE_CURL_CONSTANT(CURLOPT_CRLF)
DEFINE_CURL_CONSTANT(CURLOPT_QUOTE)
DEFINE_CURL_CONSTANT(CURLOPT_HEADERDATA)
DEFINE_CURL_CONSTANT(CURLOPT_COOKIEFILE)
DEFINE_CURL_CONSTANT(CURLOPT_SSLVERSION)
DEFINE_CURL_CONSTANT(CURLOPT_TIMECONDITION)
DEFINE_CURL_CONSTANT(CURLOPT_TIMEVALUE)
DEFINE_CURL_CONSTANT(CURLOPT_CUSTOMREQUEST)
DEFINE_CURL_CONSTANT(CURLOPT_STDERR)
DEFINE_CURL_CONSTANT(CURLOPT_POSTQUOTE)
DEFINE_CURL_CONSTANT(CURLOPT_VERBOSE)
DEFINE_CURL_CONSTANT(CURLOPT_HEADER)
DEFINE_CURL_CONSTANT(CURLOPT_NOPROGRESS)
DEFINE_CURL_CONSTANT(CURLOPT_NOBODY)
DEFINE_CURL_CONSTANT(CURLOPT_FAILONERROR)
DEFINE_CURL_CONSTANT(CURLOPT_UPLOAD)
DEFINE_CURL_CONSTANT(CURLOPT_POST)
DEFINE_CURL_CONSTANT(CURLOPT_DIRLISTONLY)
DEFINE_CURL_CONSTANT(CURLOPT_APPEND)
DEFINE_CURL_CONSTANT(CURLOPT_NETRC)
DEFINE_CURL_CONSTANT(CURLOPT_FOLLOWLOCATION)
DEFINE_CURL_CONSTANT(CURLOPT_TRANSFERTEXT)
DEFINE_CURL_CONSTANT(CURLOPT_PUT)
DEFINE_CURL_CONSTANT(CURLOPT_AUTOREFERER)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYPORT)
DEFINE_CURL_CONSTANT(CURLOPT_POSTFIELDSIZE)
DEFINE_CURL_CONSTANT(CURLOPT_HTTPPROXYTUNNEL)
DEFINE_CURL_CONSTANT(CURLOPT_INTERFACE)
DEFINE_CURL_CONSTANT(CURLOPT_KRBLEVEL)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_VERIFYPEER)
DEFINE_CURL_CONSTANT(CURLOPT_CAINFO)
DEFINE_CURL_CONSTANT(CURLOPT_MAXREDIRS)
DEFINE_CURL_CONSTANT(CURLOPT_FILETIME)
DEFINE_CURL_CONSTANT(CURLOPT_TELNETOPTIONS)
DEFINE_CURL_CONSTANT(CURLOPT_MAXCONNECTS)
DEFINE_CURL_CONSTANT(CURLOPT_FRESH_CONNECT)
DEFINE_CURL_CONSTANT(CURLOPT_FORBID_REUSE)
DEFINE_CURL_CONSTANT(CURLOPT_RANDOM_FILE)
DEFINE_CURL_CONSTANT(CURLOPT_EGDSOCKET)
DEFINE_CURL_CONSTANT(CURLOPT_CONNECTTIMEOUT)
DEFINE_CURL_CONSTANT(CURLOPT_HTTPGET)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_VERIFYHOST)
DEFINE_CURL_CONSTANT(CURLOPT_COOKIEJAR)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_CIPHER_LIST)
DEFINE_CURL_CONSTANT(CURLOPT_HTTP_VERSION)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_USE_EPSV)
DEFINE_CURL_CONSTANT(CURLOPT_SSLCERTTYPE)
DEFINE_CURL_CONSTANT(CURLOPT_SSLKEY)
DEFINE_CURL_CONSTANT(CURLOPT_SSLKEYTYPE)
DEFINE_CURL_CONSTANT(CURLOPT_SSLENGINE)
DEFINE_CURL_CONSTANT(CURLOPT_SSLENGINE_DEFAULT)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_USE_GLOBAL_CACHE)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_CACHE_TIMEOUT)
DEFINE_CURL_CONSTANT(CURLOPT_PREQUOTE)
DEFINE_CURL_CONSTANT(CURLOPT_COOKIESESSION)
DEFINE_CURL_CONSTANT(CURLOPT_CAPATH)
DEFINE_CURL_CONSTANT(CURLOPT_BUFFERSIZE)
DEFINE_CURL_CONSTANT(CURLOPT_NOSIGNAL)
DEFINE_CURL_CONSTANT(CURLOPT_SHARE)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYTYPE)
DEFINE_CURL_CONSTANT(CURLOPT_ACCEPT_ENCODING)
DEFINE_CURL_CONSTANT(CURLOPT_PRIVATE)
DEFINE_CURL_CONSTANT(CURLOPT_HTTP200ALIASES)
DEFINE_CURL_CONSTANT(CURLOPT_HTTP09_ALLOWED)
DEFINE_CURL_CONSTANT(CURLOPT_UNRESTRICTED_AUTH)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_USE_EPRT)
DEFINE_CURL_CONSTANT(CURLOPT_HTTPAUTH)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_CREATE_MISSING_DIRS)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYAUTH)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_RESPONSE_TIMEOUT)
DEFINE_CURL_CONSTANT(CURLOPT_SERVER_RESPONSE_TIMEOUT)
DEFINE_CURL_CONSTANT(CURLOPT_IPRESOLVE)
DEFINE_CURL_CONSTANT(CURLOPT_MAXFILESIZE)
DEFINE_CURL_CONSTANT(CURLOPT_INFILESIZE_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_RESUME_FROM_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_MAXFILESIZE_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_NETRC_FILE)
DEFINE_CURL_CONSTANT(CURLOPT_USE_SSL)
DEFINE_CURL_CONSTANT(CURLOPT_POSTFIELDSIZE_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_TCP_NODELAY)
DEFINE_CURL_CONSTANT(CURLOPT_FTPSSLAUTH)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_ACCOUNT)
DEFINE_CURL_CONSTANT(CURLOPT_COOKIELIST)
DEFINE_CURL_CONSTANT(CURLOPT_IGNORE_CONTENT_LENGTH)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_SKIP_PASV_IP)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_FILEMETHOD)
DEFINE_CURL_CONSTANT(CURLOPT_LOCALPORT)
DEFINE_CURL_CONSTANT(CURLOPT_LOCALPORTRANGE)
DEFINE_CURL_CONSTANT(CURLOPT_CONNECT_ONLY)
DEFINE_CURL_CONSTANT(CURLOPT_MAX_SEND_SPEED_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_MAX_RECV_SPEED_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_ALTERNATIVE_TO_USER)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_SESSIONID_CACHE)
DEFINE_CURL_CONSTANT(CURLOPT_SSH_AUTH_TYPES)
DEFINE_CURL_CONSTANT(CURLOPT_SSH_PUBLIC_KEYFILE)
DEFINE_CURL_CONSTANT(CURLOPT_SSH_PRIVATE_KEYFILE)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_SSL_CCC)
DEFINE_CURL_CONSTANT(CURLOPT_TIMEOUT_MS)
DEFINE_CURL_CONSTANT(CURLOPT_CONNECTTIMEOUT_MS)
DEFINE_CURL_CONSTANT(CURLOPT_HTTP_TRANSFER_DECODING)
DEFINE_CURL_CONSTANT(CURLOPT_HTTP_CONTENT_DECODING)
DEFINE_CURL_CONSTANT(CURLOPT_NEW_FILE_PERMS)
DEFINE_CURL_CONSTANT(CURLOPT_NEW_DIRECTORY_PERMS)
DEFINE_CURL_CONSTANT(CURLOPT_POSTREDIR)
DEFINE_CURL_CONSTANT(CURLOPT_SSH_HOST_PUBLIC_KEY_MD5)
DEFINE_CURL_CONSTANT(CURLOPT_COPYPOSTFIELDS)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_TRANSFER_MODE)
DEFINE_CURL_CONSTANT(CURLOPT_CRLFILE)
DEFINE_CURL_CONSTANT(CURLOPT_ISSUERCERT)
DEFINE_CURL_CONSTANT(CURLOPT_ADDRESS_SCOPE)
DEFINE_CURL_CONSTANT(CURLOPT_CERTINFO)
DEFINE_CURL_CONSTANT(CURLOPT_USERNAME)
DEFINE_CURL_CONSTANT(CURLOPT_PASSWORD)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYUSERNAME)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYPASSWORD)
DEFINE_CURL_CONSTANT(CURLOPT_NOPROXY)
DEFINE_CURL_CONSTANT(CURLOPT_TFTP_BLKSIZE)
DEFINE_CURL_CONSTANT(CURLOPT_SOCKS5_GSSAPI_NEC)
DEFINE_CURL_CONSTANT(CURLOPT_PROTOCOLS)
DEFINE_CURL_CONSTANT(CURLOPT_REDIR_PROTOCOLS)
DEFINE_CURL_CONSTANT(CURLOPT_SSH_KNOWNHOSTS)
DEFINE_CURL_CONSTANT(CURLOPT_MAIL_FROM)
DEFINE_CURL_CONSTANT(CURLOPT_MAIL_RCPT)
DEFINE_CURL_CONSTANT(CURLOPT_FTP_USE_PRET)
DEFINE_CURL_CONSTANT(CURLOPT_RTSP_REQUEST)
DEFINE_CURL_CONSTANT(CURLOPT_RTSP_SESSION_ID)
DEFINE_CURL_CONSTANT(CURLOPT_RTSP_STREAM_URI)
DEFINE_CURL_CONSTANT(CURLOPT_RTSP_TRANSPORT)
DEFINE_CURL_CONSTANT(CURLOPT_RTSP_CLIENT_CSEQ)
DEFINE_CURL_CONSTANT(CURLOPT_RTSP_SERVER_CSEQ)
DEFINE_CURL_CONSTANT(CURLOPT_WILDCARDMATCH)
DEFINE_CURL_CONSTANT(CURLOPT_CHUNK_DATA)
DEFINE_CURL_CONSTANT(CURLOPT_FNMATCH_DATA)
DEFINE_CURL_CONSTANT(CURLOPT_RESOLVE)
DEFINE_CURL_CONSTANT(CURLOPT_TLSAUTH_USERNAME)
DEFINE_CURL_CONSTANT(CURLOPT_TLSAUTH_PASSWORD)
DEFINE_CURL_CONSTANT(CURLOPT_TLSAUTH_TYPE)
DEFINE_CURL_CONSTANT(CURLOPT_TRANSFER_ENCODING)
DEFINE_CURL_CONSTANT(CURLOPT_GSSAPI_DELEGATION)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_SERVERS)
DEFINE_CURL_CONSTANT(CURLOPT_ACCEPTTIMEOUT_MS)
DEFINE_CURL_CONSTANT(CURLOPT_TCP_KEEPALIVE)
DEFINE_CURL_CONSTANT(CURLOPT_TCP_KEEPIDLE)
DEFINE_CURL_CONSTANT(CURLOPT_TCP_KEEPINTVL)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_OPTIONS)
DEFINE_CURL_CONSTANT(CURLOPT_MAIL_AUTH)
DEFINE_CURL_CONSTANT(CURLOPT_SASL_IR)
DEFINE_CURL_CONSTANT(CURLOPT_XOAUTH2_BEARER)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_INTERFACE)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_LOCAL_IP4)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_LOCAL_IP6)
DEFINE_CURL_CONSTANT(CURLOPT_LOGIN_OPTIONS)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_ENABLE_NPN)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_ENABLE_ALPN)
DEFINE_CURL_CONSTANT(CURLOPT_EXPECT_100_TIMEOUT_MS)
DEFINE_CURL_CONSTANT(CURLOPT_PROXYHEADER)
DEFINE_CURL_CONSTANT(CURLOPT_HEADEROPT)
DEFINE_CURL_CONSTANT(CURLOPT_PINNEDPUBLICKEY)
DEFINE_CURL_CONSTANT(CURLOPT_UNIX_SOCKET_PATH)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_VERIFYSTATUS)
DEFINE_CURL_CONSTANT(CURLOPT_SSL_FALSESTART)
DEFINE_CURL_CONSTANT(CURLOPT_PATH_AS_IS)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SERVICE_NAME)
DEFINE_CURL_CONSTANT(CURLOPT_SERVICE_NAME)
DEFINE_CURL_CONSTANT(CURLOPT_PIPEWAIT)
DEFINE_CURL_CONSTANT(CURLOPT_DEFAULT_PROTOCOL)
DEFINE_CURL_CONSTANT(CURLOPT_STREAM_WEIGHT)
DEFINE_CURL_CONSTANT(CURLOPT_STREAM_DEPENDS)
DEFINE_CURL_CONSTANT(CURLOPT_STREAM_DEPENDS_E)
DEFINE_CURL_CONSTANT(CURLOPT_TFTP_NO_OPTIONS)
DEFINE_CURL_CONSTANT(CURLOPT_CONNECT_TO)
DEFINE_CURL_CONSTANT(CURLOPT_TCP_FASTOPEN)
DEFINE_CURL_CONSTANT(CURLOPT_KEEP_SENDING_ON_ERROR)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_CAINFO)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_CAPATH)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSL_VERIFYPEER)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSL_VERIFYHOST)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSLVERSION)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_TLSAUTH_USERNAME)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_TLSAUTH_PASSWORD)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_TLSAUTH_TYPE)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSLCERT)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSLCERTTYPE)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSLKEY)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSLKEYTYPE)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_KEYPASSWD)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSL_CIPHER_LIST)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_CRLFILE)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_SSL_OPTIONS)
DEFINE_CURL_CONSTANT(CURLOPT_PRE_PROXY)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_PINNEDPUBLICKEY)
DEFINE_CURL_CONSTANT(CURLOPT_ABSTRACT_UNIX_SOCKET)
DEFINE_CURL_CONSTANT(CURLOPT_SUPPRESS_CONNECT_HEADERS)
DEFINE_CURL_CONSTANT(CURLOPT_REQUEST_TARGET)
DEFINE_CURL_CONSTANT(CURLOPT_SOCKS5_AUTH)
DEFINE_CURL_CONSTANT(CURLOPT_SSH_COMPRESSION)
DEFINE_CURL_CONSTANT(CURLOPT_MIMEPOST)
DEFINE_CURL_CONSTANT(CURLOPT_TIMEVALUE_LARGE)
DEFINE_CURL_CONSTANT(CURLOPT_HAPPY_EYEBALLS_TIMEOUT_MS)
DEFINE_CURL_CONSTANT(CURLOPT_HAPROXYPROTOCOL)
DEFINE_CURL_CONSTANT(CURLOPT_DNS_SHUFFLE_ADDRESSES)
DEFINE_CURL_CONSTANT(CURLOPT_TLS13_CIPHERS)
DEFINE_CURL_CONSTANT(CURLOPT_PROXY_TLS13_CIPHERS)
DEFINE_CURL_CONSTANT(CURLOPT_DISALLOW_USERNAME_IN_URL)
DEFINE_CURL_CONSTANT(CURLOPT_DOH_URL)
DEFINE_CURL_CONSTANT(CURLOPT_UPLOAD_BUFFERSIZE)
DEFINE_CURL_CONSTANT(CURLOPT_UPKEEP_INTERVAL_MS)
DEFINE_CURL_CONSTANT(CURLOPT_CURLU)
DEFINE_CURL_CONSTANT(CURLOPT_ALTSVC_CTRL)
DEFINE_CURL_CONSTANT(CURLOPT_ALTSVC)
DEFINE_CURL_CONSTANT(CURLOPT_MAXAGE_CONN)
DEFINE_CURL_CONSTANT(CURLOPT_SASL_AUTHZID)

// CURL INFO
DEFINE_CURL_CONSTANT(CURLINFO_EFFECTIVE_URL)
DEFINE_CURL_CONSTANT(CURLINFO_RESPONSE_CODE)
DEFINE_CURL_CONSTANT(CURLINFO_TOTAL_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_NAMELOOKUP_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_CONNECT_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_PRETRANSFER_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_SIZE_UPLOAD)
DEFINE_CURL_CONSTANT(CURLINFO_SIZE_UPLOAD_T)
DEFINE_CURL_CONSTANT(CURLINFO_SIZE_DOWNLOAD)
DEFINE_CURL_CONSTANT(CURLINFO_SIZE_DOWNLOAD_T)
DEFINE_CURL_CONSTANT(CURLINFO_SPEED_DOWNLOAD)
DEFINE_CURL_CONSTANT(CURLINFO_SPEED_DOWNLOAD_T)
DEFINE_CURL_CONSTANT(CURLINFO_SPEED_UPLOAD)
DEFINE_CURL_CONSTANT(CURLINFO_SPEED_UPLOAD_T)
DEFINE_CURL_CONSTANT(CURLINFO_HEADER_SIZE)
DEFINE_CURL_CONSTANT(CURLINFO_REQUEST_SIZE)
DEFINE_CURL_CONSTANT(CURLINFO_SSL_VERIFYRESULT)
DEFINE_CURL_CONSTANT(CURLINFO_FILETIME)
DEFINE_CURL_CONSTANT(CURLINFO_FILETIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_DOWNLOAD)
DEFINE_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_DOWNLOAD_T)
DEFINE_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_UPLOAD)
DEFINE_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_UPLOAD_T)
DEFINE_CURL_CONSTANT(CURLINFO_STARTTRANSFER_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_CONTENT_TYPE)
DEFINE_CURL_CONSTANT(CURLINFO_REDIRECT_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_REDIRECT_COUNT)
DEFINE_CURL_CONSTANT(CURLINFO_HTTP_CONNECTCODE)
DEFINE_CURL_CONSTANT(CURLINFO_HTTPAUTH_AVAIL)
DEFINE_CURL_CONSTANT(CURLINFO_PROXYAUTH_AVAIL)
DEFINE_CURL_CONSTANT(CURLINFO_OS_ERRNO)
DEFINE_CURL_CONSTANT(CURLINFO_NUM_CONNECTS)
DEFINE_CURL_CONSTANT(CURLINFO_COOKIELIST)
DEFINE_CURL_CONSTANT(CURLINFO_FTP_ENTRY_PATH)
DEFINE_CURL_CONSTANT(CURLINFO_REDIRECT_URL)
DEFINE_CURL_CONSTANT(CURLINFO_PRIMARY_IP)
DEFINE_CURL_CONSTANT(CURLINFO_APPCONNECT_TIME)
DEFINE_CURL_CONSTANT(CURLINFO_CERTINFO)
DEFINE_CURL_CONSTANT(CURLINFO_CONDITION_UNMET)
DEFINE_CURL_CONSTANT(CURLINFO_RTSP_SESSION_ID)
DEFINE_CURL_CONSTANT(CURLINFO_RTSP_CLIENT_CSEQ)
DEFINE_CURL_CONSTANT(CURLINFO_RTSP_SERVER_CSEQ)
DEFINE_CURL_CONSTANT(CURLINFO_RTSP_CSEQ_RECV)
DEFINE_CURL_CONSTANT(CURLINFO_PRIMARY_PORT)
DEFINE_CURL_CONSTANT(CURLINFO_LOCAL_IP)
DEFINE_CURL_CONSTANT(CURLINFO_LOCAL_PORT)
DEFINE_CURL_CONSTANT(CURLINFO_HTTP_VERSION)
DEFINE_CURL_CONSTANT(CURLINFO_PROXY_SSL_VERIFYRESULT)
DEFINE_CURL_CONSTANT(CURLINFO_PROTOCOL)
DEFINE_CURL_CONSTANT(CURLINFO_SCHEME)
DEFINE_CURL_CONSTANT(CURLINFO_TOTAL_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_NAMELOOKUP_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_CONNECT_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_PRETRANSFER_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_STARTTRANSFER_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_REDIRECT_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_APPCONNECT_TIME_T)
DEFINE_CURL_CONSTANT(CURLINFO_RETRY_AFTER)

// CURL FILE TYPES
DEFINE_CURL_CONSTANT(CURLFILETYPE_FILE)
DEFINE_CURL_CONSTANT(CURLFILETYPE_DIRECTORY)
DEFINE_CURL_CONSTANT(CURLFILETYPE_SYMLINK)
DEFINE_CURL_CONSTANT(CURLFILETYPE_DEVICE_CHAR)
DEFINE_CURL_CONSTANT(CURLFILETYPE_NAMEDPIPE)
DEFINE_CURL_CONSTANT(CURLFILETYPE_SOCKET)
DEFINE_CURL_CONSTANT(CURLFILETYPE_DOOR)

// CURL AUTH VALUES
DEFINE_CURL_CONSTANT(CURLAUTH_NONE)
DEFINE_CURL_CONSTANT(CURLAUTH_BASIC)
DEFINE_CURL_CONSTANT(CURLAUTH_DIGEST)
DEFINE_CURL_CONSTANT(CURLAUTH_NEGOTIATE)
DEFINE_CURL_CONSTANT(CURLAUTH_GSSNEGOTIATE)
DEFINE_CURL_CONSTANT(CURLAUTH_NTLM)
DEFINE_CURL_CONSTANT(CURLAUTH_DIGEST_IE)
DEFINE_CURL_CONSTANT(CURLAUTH_NTLM_WB)
DEFINE_CURL_CONSTANT(CURLAUTH_BEARER)
DEFINE_CURL_CONSTANT(CURLAUTH_ONLY)
DEFINE_CURL_CONSTANT(CURLAUTH_ANY)
DEFINE_CURL_CONSTANT(CURLAUTH_ANYSAFE)
//ssh
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_ANY)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_NONE)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_PUBLICKEY)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_PASSWORD)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_HOST)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_KEYBOARD)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_AGENT)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_GSSAPI)
DEFINE_CURL_CONSTANT(CURLSSH_AUTH_DEFAULT)

// usessl
DEFINE_CURL_CONSTANT(CURLUSESSL_NONE)
DEFINE_CURL_CONSTANT(CURLUSESSL_TRY)
DEFINE_CURL_CONSTANT(CURLUSESSL_CONTROL)
DEFINE_CURL_CONSTANT(CURLUSESSL_ALL)

const char *get_error_message(int result) {
  switch(result) {
    case CURLE_UNSUPPORTED_PROTOCOL: return "unsupported protocol";
    case CURLE_FAILED_INIT: return "library init failed";
    case CURLE_URL_MALFORMAT: return "url malformed";
    case CURLE_COULDNT_RESOLVE_PROXY: return "proxy resolution failed";
    case CURLE_COULDNT_RESOLVE_HOST: return "host resolution failed";
    case CURLE_COULDNT_CONNECT: return "connection failed";
    case CURLE_WEIRD_SERVER_REPLY: return "server response not understood";
    case CURLE_REMOTE_ACCESS_DENIED: return "remote access denied";
    case CURLE_FTP_ACCEPT_FAILED: return "ftp accept failed";
    case CURLE_FTP_WEIRD_PASS_REPLY: return "ftp weird pass reply";
    case CURLE_FTP_ACCEPT_TIMEOUT: return "ftp accept timeout";
    case CURLE_FTP_WEIRD_PASV_REPLY: return "ftp weird pasv reply";
    case CURLE_FTP_WEIRD_227_FORMAT: return "ftp weird 227_format";
    case CURLE_FTP_CANT_GET_HOST: return "ftp cant get host";
    case CURLE_FTP_COULDNT_SET_TYPE: return "ftp couldn't set type";
    case CURLE_PARTIAL_FILE: return "partial file";
    case CURLE_FTP_COULDNT_RETR_FILE: return "ftp couldn't retr file";
    case CURLE_QUOTE_ERROR: return "quote error";
    case CURLE_HTTP_RETURNED_ERROR: return "http returned error";
    case CURLE_WRITE_ERROR: return "write error";
    case CURLE_UPLOAD_FAILED: return "upload failed";
    case CURLE_READ_ERROR: return "read error";
    case CURLE_OUT_OF_MEMORY: return "out of memory";
    case CURLE_OPERATION_TIMEDOUT: return "operation timed out";
    case CURLE_FTP_PORT_FAILED: return "ftp port failed";
    case CURLE_FTP_COULDNT_USE_REST: return "ftp couldn't use rest";
    case CURLE_RANGE_ERROR: return "range error";
    case CURLE_HTTP_POST_ERROR: return "http post error";
    case CURLE_SSL_CONNECT_ERROR: return "ssl connect error";
    case CURLE_BAD_DOWNLOAD_RESUME: return "bad download resume";
    case CURLE_FILE_COULDNT_READ_FILE: return "file couldn't read file";
    case CURLE_LDAP_CANNOT_BIND: return "ldap cannot bind";
    case CURLE_LDAP_SEARCH_FAILED: return "ldap search failed";
    case CURLE_FUNCTION_NOT_FOUND: return "function not found";
    case CURLE_ABORTED_BY_CALLBACK: return "aborted by callback";
    case CURLE_BAD_FUNCTION_ARGUMENT: return "invalid argument";
    case CURLE_INTERFACE_FAILED: return "interface failed";
    case CURLE_TOO_MANY_REDIRECTS: return "too many redirects";
    case CURLE_UNKNOWN_OPTION: return "unknown option";
    case CURLE_GOT_NOTHING: return "got nothing";
    case CURLE_SSL_ENGINE_NOTFOUND: return "ssl engine notfound";
    case CURLE_SSL_ENGINE_SETFAILED: return "ssl engine set failed";
    case CURLE_SEND_ERROR: return "send error";
    case CURLE_RECV_ERROR: return "receive error";
    case CURLE_SSL_CERTPROBLEM: return "ssl certificate problem";
    case CURLE_SSL_CIPHER: return "ssl cipher";
    case CURLE_PEER_FAILED_VERIFICATION: return "peer failed verification";
    case CURLE_BAD_CONTENT_ENCODING: return "bad content encoding";
    case CURLE_LDAP_INVALID_URL: return "ldap invalid url";
    case CURLE_FILESIZE_EXCEEDED: return "filesize exceeded";
    case CURLE_USE_SSL_FAILED: return "use ssl failed";
    case CURLE_SEND_FAIL_REWIND: return "send fail rewind";
    case CURLE_SSL_ENGINE_INITFAILED: return "ssl engine init failed";
    case CURLE_LOGIN_DENIED: return "login denied";
    case CURLE_TFTP_NOTFOUND: return "tftp not found";
    case CURLE_TFTP_PERM: return "tftp perm";
    case CURLE_REMOTE_DISK_FULL: return "remote disk full";
    case CURLE_TFTP_ILLEGAL: return "tftp illegal";
    case CURLE_TFTP_UNKNOWNID: return "tftp unknown id";
    case CURLE_REMOTE_FILE_EXISTS: return "remote file exists";
    case CURLE_TFTP_NOSUCHUSER: return "tftp no such user";
    case CURLE_CONV_FAILED: return "conversion failed";
    case CURLE_CONV_REQD: return "conversion required";
    case CURLE_SSL_CACERT_BADFILE: return "ssl CA cert bad file";
    case CURLE_REMOTE_FILE_NOT_FOUND: return "remote file not found";
    case CURLE_SSL_SHUTDOWN_FAILED: return "ssl shutdown failed";
    case CURLE_SSL_CRL_BADFILE: return "ssl crl bad file";
    case CURLE_SSL_ISSUER_ERROR: return "ssl issuer error";
    case CURLE_FTP_PRET_FAILED: return "ftp PRET failed";
    case CURLE_RTSP_CSEQ_ERROR: return "rtsp CSEQ error";
    case CURLE_RTSP_SESSION_ERROR: return "rtsp session error";
    case CURLE_FTP_BAD_FILE_LIST: return "ftp bad file list";
    case CURLE_CHUNK_FAILED: return "chunk failed";
    case CURLE_NO_CONNECTION_AVAILABLE: return "no connection available";
    case CURLE_SSL_PINNEDPUBKEYNOTMATCH: return "ssl pinned public key does not match";
    case CURLE_SSL_INVALIDCERTSTATUS: return "ssl invalid cert status";
    case CURLE_AUTH_ERROR: return "auth error";
    default: return "error occurred";
  }
}

struct b_curl_string {
  unsigned char *ptr;
  size_t len;
};

void init_string(struct b_curl_string *s) {
  s->len = 0;
  s->ptr = malloc(s->len+1);
  if (s->ptr == NULL) {
    fprintf(stderr, "device out of memory\n");
    exit(EXIT_FAILURE);
  }
  s->ptr[0] = '\0';
}

size_t b_CurlWriteFunc(void *ptr, size_t size, size_t nmemb, struct b_curl_string *s) {
  size_t new_len = s->len + size*nmemb;
  s->ptr = realloc(s->ptr, new_len+1);
  if (s->ptr == NULL) {
    fprintf(stderr, "device out of memory\n");
    exit(EXIT_FAILURE);
  }
  memcpy(s->ptr+s->len, ptr, size*nmemb);
  s->ptr[new_len] = '\0';
  s->len = new_len;

  return size*nmemb;
}

b_value __curl_version(b_vm *vm) {
  char *version = curl_version();
  return STRING_VAL(version);
}

DECLARE_MODULE_METHOD(curl__easy_init) {
  ENFORCE_ARG_COUNT(easy_init, 0);
  CURL *curl = curl_easy_init();
  if(curl != NULL) {
    RETURN_PTR(curl);
  }
  RETURN_ERROR("Error creating new instance of Curl");
}

DECLARE_MODULE_METHOD(curl__easy_cleanup) {
  ENFORCE_ARG_COUNT(easy_cleanup, 1);
  ENFORCE_ARG_TYPE(easy_cleanup, 0, IS_PTR);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  curl_easy_cleanup(curl);
  RETURN;
}

DECLARE_MODULE_METHOD(curl__easy_reset) {
  ENFORCE_ARG_COUNT(easy_reset, 1);
  ENFORCE_ARG_TYPE(easy_reset, 0, IS_PTR);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  curl_easy_reset(curl);
  RETURN;
}

DECLARE_MODULE_METHOD(curl__easy_setopt) {
  ENFORCE_ARG_COUNT(easy_setopt, 3);
  ENFORCE_ARG_TYPE(easy_setopt, 0, IS_PTR);
  ENFORCE_ARG_TYPE(easy_setopt, 1, IS_NUMBER);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  int opt = AS_NUMBER(args[1]);

  int result = CURLE_BAD_FUNCTION_ARGUMENT;
  if(IS_STRING(args[2])) {
    b_obj_string *str = AS_STRING(args[2]);
    if(str->length > CURL_MAX_INPUT_LENGTH) {
      RETURN_ERROR("string length exceeds maximum input length of %d", CURL_MAX_INPUT_LENGTH);
    }

    result = curl_easy_setopt(curl, opt, str->chars);
  } else if(IS_BYTES(args[2])) {
    b_obj_bytes *bytes = AS_BYTES(args[2]);
    if(bytes->bytes.count > CURL_MAX_INPUT_LENGTH) {
      RETURN_ERROR("bytes length exceeds maximum input length of %d", CURL_MAX_INPUT_LENGTH);
    }

    result = curl_easy_setopt(curl, opt, bytes->bytes.bytes);
  } else if(IS_NUMBER(args[2])) {
    result = curl_easy_setopt(curl, opt, (long)AS_NUMBER(args[2]));
  } else if(IS_BOOL(args[2])) {
    result = curl_easy_setopt(curl, opt, AS_BOOL(args[2]) ? 1L : 0L);
  } else if(IS_PTR(args[2])) {
    result = curl_easy_setopt(curl, opt, AS_PTR(args[2])->pointer);
  } else if(IS_NIL(args[2])) {
    result = curl_easy_setopt(curl, opt, NULL);
  }

  if(result == CURLE_OK) {
    RETURN_TRUE;
  }

  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__easy_perform) {
  ENFORCE_ARG_COUNT(easy_perform, 1);
  ENFORCE_ARG_TYPE(easy_perform, 0, IS_PTR);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;

  // @TODO: set write, header and debug functions here...
  struct b_curl_string body;
  init_string(&body);
  curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, b_CurlWriteFunc);
  curl_easy_setopt(curl, CURLOPT_WRITEDATA, &body);

  struct b_curl_string headers;
  init_string(&headers);
  curl_easy_setopt(curl, CURLOPT_HEADERFUNCTION, b_CurlWriteFunc);
  curl_easy_setopt(curl, CURLOPT_HEADERDATA, &headers);

  int result = curl_easy_perform(curl);
  if(result == CURLE_OK) {

    b_obj_dict *dict = (b_obj_dict*)GC(new_dict(vm));
    b_obj_bytes *data = (b_obj_bytes*)GC(copy_bytes(vm, body.ptr, body.len));

    dict_add_entry(vm, dict, GC_L_STRING("headers", 7), GC_STRING((char *)headers.ptr));
    dict_add_entry(vm, dict, GC_L_STRING("body", 4), OBJ_VAL(data));

    free(body.ptr);
    free(headers.ptr);

    RETURN_OBJ(dict);
  } else {
    free(body.ptr);
    free(headers.ptr);
    RETURN_ERROR(get_error_message(result));
  }
}

DECLARE_MODULE_METHOD(curl__easy_getinfo) {
  ENFORCE_ARG_COUNT(easy_getinfo, 2);
  ENFORCE_ARG_TYPE(easy_getinfo, 0, IS_PTR);
  ENFORCE_ARG_TYPE(easy_getinfo, 1, IS_NUMBER);

  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  int info = AS_NUMBER(args[1]);

  bool found = false;
  for(int i = 0; i < 58; i++) {
    if(CurlInfoTypes[i] == info)
      found = true;
  }

  if(!found) {
    RETURN_ERROR("invalid info type");
  }

  int result = CURLE_BAD_FUNCTION_ARGUMENT;
  switch(CURLInfoTypeMap[info].type) {
    case CURLINFO_STRING: {
      char *data = NULL;
      result = curl_easy_getinfo(curl, info, &data);
      if(result == CURLE_OK) {
        RETURN_STRING(data);
      }
      break;
    }
    case CURLINFO_LONG: {
      long data = 0L;
      result = curl_easy_getinfo(curl, info, &data);
      if(result == CURLE_OK) {

        if(info == CURLINFO_PROTOCOL) {
          if(data != -1) {
            RETURN_STRING(CURLProtoMap[(int)log2(data)]);
          } else {
            RETURN_STRING("all");
          }
        }

        RETURN_NUMBER(data);
      }
      break;
    }
    case CURLINFO_DOUBLE: {
      double data = 0.0;
      result = curl_easy_getinfo(curl, info, &data);
      if(result == CURLE_OK) {
        RETURN_NUMBER(data);
      }
      break;
    }
    case CURLINFO_OFF_T: {
      curl_off_t data;
      result = curl_easy_getinfo(curl, info, &data);
      if(result == CURLE_OK) {
        RETURN_NUMBER(data);
      }
      break;
    }
    case CURLINFO_SLIST: {
      if(info == CURLINFO_CERTINFO) {
        struct curl_certinfo *ci;
        result = curl_easy_getinfo(curl, CURLINFO_CERTINFO, &ci);

        if (result == CURLE_OK) {
          b_obj_list *list = (b_obj_list*)GC(new_list(vm));

          for(int i = 0; i < ci->num_of_certs; i++) {
            b_obj_list *inner_list = (b_obj_list*)GC(new_list(vm));
            struct curl_slist *s_list;

            for(s_list = ci->certinfo[i]; s_list; s_list = s_list->next)
              write_list(vm, inner_list, GC_STRING(s_list->data));

            write_list(vm, list, OBJ_VAL(inner_list));
          }

          RETURN_OBJ(list);
        }
      } else {
        struct curl_slist *data = NULL;
        result = curl_easy_getinfo(curl, info, data);
        if(result == CURLE_OK) {
          b_obj_list *list = (b_obj_list*)GC(new_list(vm));
          while(data) {
            write_list(vm, list, GC_STRING(data->data));
            data = data->next;
          }
          curl_slist_free_all(data);
          RETURN_OBJ(list);
        }
      }
      break;
    }
    default: {
      break;
    }
  }

  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__easy_escape) {
  ENFORCE_ARG_COUNT(easy_escape, 2);
  ENFORCE_ARG_TYPE(easy_escape, 0, IS_PTR);
  ENFORCE_ARG_TYPE(easy_escape, 1, IS_STRING);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  b_obj_string *string = AS_STRING(args[1]);
  char *result = curl_easy_escape(curl, string->chars, string->length);

  if(result != NULL) {
    b_obj_string *n_string = (b_obj_string*)GC(copy_string(vm, result, (int)strlen(result)));
    curl_free(result);
    RETURN_OBJ(n_string);
  }

  RETURN_NIL;
}

DECLARE_MODULE_METHOD(curl__easy_unescape) {
  ENFORCE_ARG_COUNT(easy_unescape, 2);
  ENFORCE_ARG_TYPE(easy_unescape, 0, IS_PTR);
  ENFORCE_ARG_TYPE(easy_unescape, 1, IS_STRING);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  b_obj_string *string = AS_STRING(args[1]);
  int out_length = 0;
  char *result = curl_easy_unescape(curl, string->chars, string->length, &out_length);

  if(result != NULL) {
    b_obj_string *n_string = (b_obj_string*)GC(copy_string(vm, result, (int)strlen(result)));
    curl_free(result);
    RETURN_OBJ(n_string);
  }

  RETURN_NIL;
}

DECLARE_MODULE_METHOD(curl__getdate) {
  ENFORCE_ARG_COUNT(getdate, 1);
  ENFORCE_ARG_TYPE(getdate, 0, IS_STRING);
  time_t tv = curl_getdate(AS_C_STRING(args[0]), NULL);
  if(tv == -1) {
    RETURN_ERROR("invalid datetime");
  }

  RETURN_NUMBER(tv);
}

DECLARE_MODULE_METHOD(curl__mime_init) {
  ENFORCE_ARG_COUNT(mime_init, 1);
  ENFORCE_ARG_TYPE(mime_init, 0, IS_PTR);
  CURL *curl = (CURL*)AS_PTR(args[0])->pointer;
  curl_mime *mime = curl_mime_init(curl);
  RETURN_PTR(mime);
}

DECLARE_MODULE_METHOD(curl__mime_addpart) {
  ENFORCE_ARG_COUNT(mime_addpart, 1);
  ENFORCE_ARG_TYPE(mime_addpart, 0, IS_PTR);
  curl_mime *mime = (curl_mime*)AS_PTR(args[0])->pointer;
  curl_mimepart *part = curl_mime_addpart(mime);
  RETURN_PTR(part);
}

DECLARE_MODULE_METHOD(curl__mime_encoding) {
  ENFORCE_ARG_COUNT(mime_encoding, 2);
  ENFORCE_ARG_TYPE(mime_encoding, 0, IS_PTR);
  ENFORCE_ARG_TYPE(mime_encoding, 1, IS_STRING);
  curl_mimepart *part = (curl_mimepart*)AS_PTR(args[0])->pointer;
  CURLcode result = curl_mime_encoder(part, AS_C_STRING(args[1]));
  if(result == CURLE_OK) {
    RETURN_TRUE;
  }
  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__mime_name) {
  ENFORCE_ARG_COUNT(mime_name, 2);
  ENFORCE_ARG_TYPE(mime_name, 0, IS_PTR);
  ENFORCE_ARG_TYPE(mime_name, 1, IS_STRING);
  curl_mimepart *part = (curl_mimepart*)AS_PTR(args[0])->pointer;
  CURLcode result = curl_mime_name(part, AS_C_STRING(args[1]));
  if(result == CURLE_OK) {
    RETURN_TRUE;
  }
  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__mime_type) {
  ENFORCE_ARG_COUNT(mime_type, 2);
  ENFORCE_ARG_TYPE(mime_type, 0, IS_PTR);
  ENFORCE_ARG_TYPE(mime_type, 1, IS_STRING);
  curl_mimepart *part = (curl_mimepart*)AS_PTR(args[0])->pointer;
  CURLcode result = curl_mime_type(part, AS_C_STRING(args[1]));
  if(result == CURLE_OK) {
    RETURN_TRUE;
  }
  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__mime_data) {
  ENFORCE_ARG_COUNT(mime_data, 2);
  ENFORCE_ARG_TYPE(mime_data, 0, IS_PTR);
  ENFORCE_ARG_TYPE(mime_data, 1, IS_STRING);
  curl_mimepart *part = (curl_mimepart*)AS_PTR(args[0])->pointer;
  CURLcode result = curl_mime_data(part, AS_C_STRING(args[1]), CURL_ZERO_TERMINATED);
  if(result == CURLE_OK) {
    RETURN_TRUE;
  }
  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__mime_filedata) {
  ENFORCE_ARG_COUNT(mime_data, 2);
  ENFORCE_ARG_TYPE(mime_data, 0, IS_PTR);
  ENFORCE_ARG_TYPE(mime_data, 1, IS_STRING);
  curl_mimepart *part = (curl_mimepart*)AS_PTR(args[0])->pointer;
  CURLcode result = curl_mime_filedata(part, AS_C_STRING(args[1]));
  if(result == CURLE_OK) {
    RETURN_TRUE;
  }
  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__mime_subparts) {
  ENFORCE_ARG_COUNT(mime_subparts, 2);
  ENFORCE_ARG_TYPE(mime_subparts, 0, IS_PTR);
  ENFORCE_ARG_TYPE(mime_subparts, 1, IS_PTR);
  curl_mimepart *part = (curl_mimepart*)AS_PTR(args[0])->pointer;
  curl_mime *subpart = (curl_mime *)AS_PTR(args[1])->pointer;
  CURLcode result = curl_mime_subparts(part, subpart);
  if(result == CURLE_OK) {
    RETURN_TRUE;
  }
  RETURN_ERROR(get_error_message(result));
}

DECLARE_MODULE_METHOD(curl__slist_create) {
  ENFORCE_ARG_COUNT(slist_create, 1);
  ENFORCE_ARG_TYPE(slist_create, 0, IS_LIST);
  b_obj_list *b_list = AS_LIST(args[0]);

  struct curl_slist *s_list = NULL;

  for(int i = 0; i < b_list->items.count; i++) {
    s_list = curl_slist_append(s_list, value_to_string(vm, b_list->items.values[i]));
  }

  if(s_list != NULL) {
    RETURN_PTR(s_list);
  }

  RETURN_NIL;
}

DECLARE_MODULE_METHOD(curl__slist_free) {
  ENFORCE_ARG_COUNT(slist_free, 1);
  ENFORCE_ARG_TYPE(slist_free, 0, IS_PTR);
  struct curl_slist *s_list = (struct curl_slist *)AS_PTR(args[0])->pointer;
  if(s_list != NULL) {
    curl_slist_free_all(s_list);
  }
  RETURN;
}

CREATE_MODULE_LOADER(curl) {

  static b_field_reg module_fields[] = {
      // CURL OPTS
      GET_CURL_CONSTANT(CURLOPT_URL),
      GET_CURL_CONSTANT(CURLOPT_PORT),
      GET_CURL_CONSTANT(CURLOPT_PROXY),
      GET_CURL_CONSTANT(CURLOPT_USERPWD),
      GET_CURL_CONSTANT(CURLOPT_PROXYUSERPWD),
      GET_CURL_CONSTANT(CURLOPT_RANGE),
      GET_CURL_CONSTANT(CURLOPT_ERRORBUFFER),
      GET_CURL_CONSTANT(CURLOPT_TIMEOUT),
      GET_CURL_CONSTANT(CURLOPT_INFILESIZE),
      GET_CURL_CONSTANT(CURLOPT_POSTFIELDS),
      GET_CURL_CONSTANT(CURLOPT_REFERER),
      GET_CURL_CONSTANT(CURLOPT_FTPPORT),
      GET_CURL_CONSTANT(CURLOPT_USERAGENT),
      GET_CURL_CONSTANT(CURLOPT_LOW_SPEED_LIMIT),
      GET_CURL_CONSTANT(CURLOPT_LOW_SPEED_TIME),
      GET_CURL_CONSTANT(CURLOPT_RESUME_FROM),
      GET_CURL_CONSTANT(CURLOPT_COOKIE),
      GET_CURL_CONSTANT(CURLOPT_HTTPHEADER),
      GET_CURL_CONSTANT(CURLOPT_HTTPPOST),
      GET_CURL_CONSTANT(CURLOPT_SSLCERT),
      GET_CURL_CONSTANT(CURLOPT_KEYPASSWD),
      GET_CURL_CONSTANT(CURLOPT_CRLF),
      GET_CURL_CONSTANT(CURLOPT_QUOTE),
      GET_CURL_CONSTANT(CURLOPT_HEADERDATA),
      GET_CURL_CONSTANT(CURLOPT_COOKIEFILE),
      GET_CURL_CONSTANT(CURLOPT_SSLVERSION),
      GET_CURL_CONSTANT(CURLOPT_TIMECONDITION),
      GET_CURL_CONSTANT(CURLOPT_TIMEVALUE),
      GET_CURL_CONSTANT(CURLOPT_CUSTOMREQUEST),
      GET_CURL_CONSTANT(CURLOPT_STDERR),
      GET_CURL_CONSTANT(CURLOPT_POSTQUOTE),
      GET_CURL_CONSTANT(CURLOPT_VERBOSE),
      GET_CURL_CONSTANT(CURLOPT_HEADER),
      GET_CURL_CONSTANT(CURLOPT_NOPROGRESS),
      GET_CURL_CONSTANT(CURLOPT_NOBODY),
      GET_CURL_CONSTANT(CURLOPT_FAILONERROR),
      GET_CURL_CONSTANT(CURLOPT_UPLOAD),
      GET_CURL_CONSTANT(CURLOPT_POST),
      GET_CURL_CONSTANT(CURLOPT_DIRLISTONLY),
      GET_CURL_CONSTANT(CURLOPT_APPEND),
      GET_CURL_CONSTANT(CURLOPT_NETRC),
      GET_CURL_CONSTANT(CURLOPT_FOLLOWLOCATION),
      GET_CURL_CONSTANT(CURLOPT_TRANSFERTEXT),
      GET_CURL_CONSTANT(CURLOPT_PUT),
      GET_CURL_CONSTANT(CURLOPT_AUTOREFERER),
      GET_CURL_CONSTANT(CURLOPT_PROXYPORT),
      GET_CURL_CONSTANT(CURLOPT_POSTFIELDSIZE),
      GET_CURL_CONSTANT(CURLOPT_HTTPPROXYTUNNEL),
      GET_CURL_CONSTANT(CURLOPT_INTERFACE),
      GET_CURL_CONSTANT(CURLOPT_KRBLEVEL),
      GET_CURL_CONSTANT(CURLOPT_SSL_VERIFYPEER),
      GET_CURL_CONSTANT(CURLOPT_CAINFO),
      GET_CURL_CONSTANT(CURLOPT_MAXREDIRS),
      GET_CURL_CONSTANT(CURLOPT_FILETIME),
      GET_CURL_CONSTANT(CURLOPT_TELNETOPTIONS),
      GET_CURL_CONSTANT(CURLOPT_MAXCONNECTS),
      GET_CURL_CONSTANT(CURLOPT_FRESH_CONNECT),
      GET_CURL_CONSTANT(CURLOPT_FORBID_REUSE),
      GET_CURL_CONSTANT(CURLOPT_RANDOM_FILE),
      GET_CURL_CONSTANT(CURLOPT_EGDSOCKET),
      GET_CURL_CONSTANT(CURLOPT_CONNECTTIMEOUT),
      GET_CURL_CONSTANT(CURLOPT_HTTPGET),
      GET_CURL_CONSTANT(CURLOPT_SSL_VERIFYHOST),
      GET_CURL_CONSTANT(CURLOPT_COOKIEJAR),
      GET_CURL_CONSTANT(CURLOPT_SSL_CIPHER_LIST),
      GET_CURL_CONSTANT(CURLOPT_HTTP_VERSION),
      GET_CURL_CONSTANT(CURLOPT_FTP_USE_EPSV),
      GET_CURL_CONSTANT(CURLOPT_SSLCERTTYPE),
      GET_CURL_CONSTANT(CURLOPT_SSLKEY),
      GET_CURL_CONSTANT(CURLOPT_SSLKEYTYPE),
      GET_CURL_CONSTANT(CURLOPT_SSLENGINE),
      GET_CURL_CONSTANT(CURLOPT_SSLENGINE_DEFAULT),
      GET_CURL_CONSTANT(CURLOPT_DNS_USE_GLOBAL_CACHE),
      GET_CURL_CONSTANT(CURLOPT_DNS_CACHE_TIMEOUT),
      GET_CURL_CONSTANT(CURLOPT_PREQUOTE),
      GET_CURL_CONSTANT(CURLOPT_COOKIESESSION),
      GET_CURL_CONSTANT(CURLOPT_CAPATH),
      GET_CURL_CONSTANT(CURLOPT_BUFFERSIZE),
      GET_CURL_CONSTANT(CURLOPT_NOSIGNAL),
      GET_CURL_CONSTANT(CURLOPT_SHARE),
      GET_CURL_CONSTANT(CURLOPT_PROXYTYPE),
      GET_CURL_CONSTANT(CURLOPT_ACCEPT_ENCODING),
      GET_CURL_CONSTANT(CURLOPT_PRIVATE),
      GET_CURL_CONSTANT(CURLOPT_HTTP200ALIASES),
      GET_CURL_CONSTANT(CURLOPT_HTTP09_ALLOWED),
      GET_CURL_CONSTANT(CURLOPT_UNRESTRICTED_AUTH),
      GET_CURL_CONSTANT(CURLOPT_FTP_USE_EPRT),
      GET_CURL_CONSTANT(CURLOPT_HTTPAUTH),
      GET_CURL_CONSTANT(CURLOPT_FTP_CREATE_MISSING_DIRS),
      GET_CURL_CONSTANT(CURLOPT_PROXYAUTH),
      GET_CURL_CONSTANT(CURLOPT_FTP_RESPONSE_TIMEOUT),
      GET_CURL_CONSTANT(CURLOPT_SERVER_RESPONSE_TIMEOUT),
      GET_CURL_CONSTANT(CURLOPT_IPRESOLVE),
      GET_CURL_CONSTANT(CURLOPT_MAXFILESIZE),
      GET_CURL_CONSTANT(CURLOPT_INFILESIZE_LARGE),
      GET_CURL_CONSTANT(CURLOPT_RESUME_FROM_LARGE),
      GET_CURL_CONSTANT(CURLOPT_MAXFILESIZE_LARGE),
      GET_CURL_CONSTANT(CURLOPT_NETRC_FILE),
      GET_CURL_CONSTANT(CURLOPT_USE_SSL),
      GET_CURL_CONSTANT(CURLOPT_POSTFIELDSIZE_LARGE),
      GET_CURL_CONSTANT(CURLOPT_TCP_NODELAY),
      GET_CURL_CONSTANT(CURLOPT_FTPSSLAUTH),
      GET_CURL_CONSTANT(CURLOPT_FTP_ACCOUNT),
      GET_CURL_CONSTANT(CURLOPT_COOKIELIST),
      GET_CURL_CONSTANT(CURLOPT_IGNORE_CONTENT_LENGTH),
      GET_CURL_CONSTANT(CURLOPT_FTP_SKIP_PASV_IP),
      GET_CURL_CONSTANT(CURLOPT_FTP_FILEMETHOD),
      GET_CURL_CONSTANT(CURLOPT_LOCALPORT),
      GET_CURL_CONSTANT(CURLOPT_LOCALPORTRANGE),
      GET_CURL_CONSTANT(CURLOPT_CONNECT_ONLY),
      GET_CURL_CONSTANT(CURLOPT_MAX_SEND_SPEED_LARGE),
      GET_CURL_CONSTANT(CURLOPT_MAX_RECV_SPEED_LARGE),
      GET_CURL_CONSTANT(CURLOPT_FTP_ALTERNATIVE_TO_USER),
      GET_CURL_CONSTANT(CURLOPT_SSL_SESSIONID_CACHE),
      GET_CURL_CONSTANT(CURLOPT_SSH_AUTH_TYPES),
      GET_CURL_CONSTANT(CURLOPT_SSH_PUBLIC_KEYFILE),
      GET_CURL_CONSTANT(CURLOPT_SSH_PRIVATE_KEYFILE),
      GET_CURL_CONSTANT(CURLOPT_FTP_SSL_CCC),
      GET_CURL_CONSTANT(CURLOPT_TIMEOUT_MS),
      GET_CURL_CONSTANT(CURLOPT_CONNECTTIMEOUT_MS),
      GET_CURL_CONSTANT(CURLOPT_HTTP_TRANSFER_DECODING),
      GET_CURL_CONSTANT(CURLOPT_HTTP_CONTENT_DECODING),
      GET_CURL_CONSTANT(CURLOPT_NEW_FILE_PERMS),
      GET_CURL_CONSTANT(CURLOPT_NEW_DIRECTORY_PERMS),
      GET_CURL_CONSTANT(CURLOPT_POSTREDIR),
      GET_CURL_CONSTANT(CURLOPT_SSH_HOST_PUBLIC_KEY_MD5),
      GET_CURL_CONSTANT(CURLOPT_COPYPOSTFIELDS),
      GET_CURL_CONSTANT(CURLOPT_PROXY_TRANSFER_MODE),
      GET_CURL_CONSTANT(CURLOPT_CRLFILE),
      GET_CURL_CONSTANT(CURLOPT_ISSUERCERT),
      GET_CURL_CONSTANT(CURLOPT_ADDRESS_SCOPE),
      GET_CURL_CONSTANT(CURLOPT_CERTINFO),
      GET_CURL_CONSTANT(CURLOPT_USERNAME),
      GET_CURL_CONSTANT(CURLOPT_PASSWORD),
      GET_CURL_CONSTANT(CURLOPT_PROXYUSERNAME),
      GET_CURL_CONSTANT(CURLOPT_PROXYPASSWORD),
      GET_CURL_CONSTANT(CURLOPT_NOPROXY),
      GET_CURL_CONSTANT(CURLOPT_TFTP_BLKSIZE),
      GET_CURL_CONSTANT(CURLOPT_SOCKS5_GSSAPI_NEC),
      GET_CURL_CONSTANT(CURLOPT_PROTOCOLS),
      GET_CURL_CONSTANT(CURLOPT_REDIR_PROTOCOLS),
      GET_CURL_CONSTANT(CURLOPT_SSH_KNOWNHOSTS),
      GET_CURL_CONSTANT(CURLOPT_MAIL_FROM),
      GET_CURL_CONSTANT(CURLOPT_MAIL_RCPT),
      GET_CURL_CONSTANT(CURLOPT_FTP_USE_PRET),
      GET_CURL_CONSTANT(CURLOPT_RTSP_REQUEST),
      GET_CURL_CONSTANT(CURLOPT_RTSP_SESSION_ID),
      GET_CURL_CONSTANT(CURLOPT_RTSP_STREAM_URI),
      GET_CURL_CONSTANT(CURLOPT_RTSP_TRANSPORT),
      GET_CURL_CONSTANT(CURLOPT_RTSP_CLIENT_CSEQ),
      GET_CURL_CONSTANT(CURLOPT_RTSP_SERVER_CSEQ),
      GET_CURL_CONSTANT(CURLOPT_WILDCARDMATCH),
      GET_CURL_CONSTANT(CURLOPT_CHUNK_DATA),
      GET_CURL_CONSTANT(CURLOPT_FNMATCH_DATA),
      GET_CURL_CONSTANT(CURLOPT_RESOLVE),
      GET_CURL_CONSTANT(CURLOPT_TLSAUTH_USERNAME),
      GET_CURL_CONSTANT(CURLOPT_TLSAUTH_PASSWORD),
      GET_CURL_CONSTANT(CURLOPT_TLSAUTH_TYPE),
      GET_CURL_CONSTANT(CURLOPT_TRANSFER_ENCODING),
      GET_CURL_CONSTANT(CURLOPT_GSSAPI_DELEGATION),
      GET_CURL_CONSTANT(CURLOPT_DNS_SERVERS),
      GET_CURL_CONSTANT(CURLOPT_ACCEPTTIMEOUT_MS),
      GET_CURL_CONSTANT(CURLOPT_TCP_KEEPALIVE),
      GET_CURL_CONSTANT(CURLOPT_TCP_KEEPIDLE),
      GET_CURL_CONSTANT(CURLOPT_TCP_KEEPINTVL),
      GET_CURL_CONSTANT(CURLOPT_SSL_OPTIONS),
      GET_CURL_CONSTANT(CURLOPT_MAIL_AUTH),
      GET_CURL_CONSTANT(CURLOPT_SASL_IR),
      GET_CURL_CONSTANT(CURLOPT_XOAUTH2_BEARER),
      GET_CURL_CONSTANT(CURLOPT_DNS_INTERFACE),
      GET_CURL_CONSTANT(CURLOPT_DNS_LOCAL_IP4),
      GET_CURL_CONSTANT(CURLOPT_DNS_LOCAL_IP6),
      GET_CURL_CONSTANT(CURLOPT_LOGIN_OPTIONS),
      GET_CURL_CONSTANT(CURLOPT_SSL_ENABLE_NPN),
      GET_CURL_CONSTANT(CURLOPT_SSL_ENABLE_ALPN),
      GET_CURL_CONSTANT(CURLOPT_EXPECT_100_TIMEOUT_MS),
      GET_CURL_CONSTANT(CURLOPT_PROXYHEADER),
      GET_CURL_CONSTANT(CURLOPT_HEADEROPT),
      GET_CURL_CONSTANT(CURLOPT_PINNEDPUBLICKEY),
      GET_CURL_CONSTANT(CURLOPT_UNIX_SOCKET_PATH),
      GET_CURL_CONSTANT(CURLOPT_SSL_VERIFYSTATUS),
      GET_CURL_CONSTANT(CURLOPT_SSL_FALSESTART),
      GET_CURL_CONSTANT(CURLOPT_PATH_AS_IS),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SERVICE_NAME),
      GET_CURL_CONSTANT(CURLOPT_SERVICE_NAME),
      GET_CURL_CONSTANT(CURLOPT_PIPEWAIT),
      GET_CURL_CONSTANT(CURLOPT_DEFAULT_PROTOCOL),
      GET_CURL_CONSTANT(CURLOPT_STREAM_WEIGHT),
      GET_CURL_CONSTANT(CURLOPT_STREAM_DEPENDS),
      GET_CURL_CONSTANT(CURLOPT_STREAM_DEPENDS_E),
      GET_CURL_CONSTANT(CURLOPT_TFTP_NO_OPTIONS),
      GET_CURL_CONSTANT(CURLOPT_CONNECT_TO),
      GET_CURL_CONSTANT(CURLOPT_TCP_FASTOPEN),
      GET_CURL_CONSTANT(CURLOPT_KEEP_SENDING_ON_ERROR),
      GET_CURL_CONSTANT(CURLOPT_PROXY_CAINFO),
      GET_CURL_CONSTANT(CURLOPT_PROXY_CAPATH),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSL_VERIFYPEER),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSL_VERIFYHOST),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSLVERSION),
      GET_CURL_CONSTANT(CURLOPT_PROXY_TLSAUTH_USERNAME),
      GET_CURL_CONSTANT(CURLOPT_PROXY_TLSAUTH_PASSWORD),
      GET_CURL_CONSTANT(CURLOPT_PROXY_TLSAUTH_TYPE),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSLCERT),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSLCERTTYPE),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSLKEY),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSLKEYTYPE),
      GET_CURL_CONSTANT(CURLOPT_PROXY_KEYPASSWD),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSL_CIPHER_LIST),
      GET_CURL_CONSTANT(CURLOPT_PROXY_CRLFILE),
      GET_CURL_CONSTANT(CURLOPT_PROXY_SSL_OPTIONS),
      GET_CURL_CONSTANT(CURLOPT_PRE_PROXY),
      GET_CURL_CONSTANT(CURLOPT_PROXY_PINNEDPUBLICKEY),
      GET_CURL_CONSTANT(CURLOPT_ABSTRACT_UNIX_SOCKET),
      GET_CURL_CONSTANT(CURLOPT_SUPPRESS_CONNECT_HEADERS),
      GET_CURL_CONSTANT(CURLOPT_REQUEST_TARGET),
      GET_CURL_CONSTANT(CURLOPT_SOCKS5_AUTH),
      GET_CURL_CONSTANT(CURLOPT_SSH_COMPRESSION),
      GET_CURL_CONSTANT(CURLOPT_MIMEPOST),
      GET_CURL_CONSTANT(CURLOPT_TIMEVALUE_LARGE),
      GET_CURL_CONSTANT(CURLOPT_HAPPY_EYEBALLS_TIMEOUT_MS),
      GET_CURL_CONSTANT(CURLOPT_HAPROXYPROTOCOL),
      GET_CURL_CONSTANT(CURLOPT_DNS_SHUFFLE_ADDRESSES),
      GET_CURL_CONSTANT(CURLOPT_TLS13_CIPHERS),
      GET_CURL_CONSTANT(CURLOPT_PROXY_TLS13_CIPHERS),
      GET_CURL_CONSTANT(CURLOPT_DISALLOW_USERNAME_IN_URL),
      GET_CURL_CONSTANT(CURLOPT_DOH_URL),
      GET_CURL_CONSTANT(CURLOPT_UPLOAD_BUFFERSIZE),
      GET_CURL_CONSTANT(CURLOPT_UPKEEP_INTERVAL_MS),
      GET_CURL_CONSTANT(CURLOPT_CURLU),
      GET_CURL_CONSTANT(CURLOPT_ALTSVC_CTRL),
      GET_CURL_CONSTANT(CURLOPT_ALTSVC),
      GET_CURL_CONSTANT(CURLOPT_MAXAGE_CONN),
      GET_CURL_CONSTANT(CURLOPT_SASL_AUTHZID),

      // CURL INFO
      GET_CURL_CONSTANT(CURLINFO_EFFECTIVE_URL),
      GET_CURL_CONSTANT(CURLINFO_RESPONSE_CODE),
      GET_CURL_CONSTANT(CURLINFO_TOTAL_TIME),
      GET_CURL_CONSTANT(CURLINFO_NAMELOOKUP_TIME),
      GET_CURL_CONSTANT(CURLINFO_CONNECT_TIME),
      GET_CURL_CONSTANT(CURLINFO_PRETRANSFER_TIME),
      GET_CURL_CONSTANT(CURLINFO_SIZE_UPLOAD),
      GET_CURL_CONSTANT(CURLINFO_SIZE_UPLOAD_T),
      GET_CURL_CONSTANT(CURLINFO_SIZE_DOWNLOAD),
      GET_CURL_CONSTANT(CURLINFO_SIZE_DOWNLOAD_T),
      GET_CURL_CONSTANT(CURLINFO_SPEED_DOWNLOAD),
      GET_CURL_CONSTANT(CURLINFO_SPEED_DOWNLOAD_T),
      GET_CURL_CONSTANT(CURLINFO_SPEED_UPLOAD),
      GET_CURL_CONSTANT(CURLINFO_SPEED_UPLOAD_T),
      GET_CURL_CONSTANT(CURLINFO_HEADER_SIZE),
      GET_CURL_CONSTANT(CURLINFO_REQUEST_SIZE),
      GET_CURL_CONSTANT(CURLINFO_SSL_VERIFYRESULT),
      GET_CURL_CONSTANT(CURLINFO_FILETIME),
      GET_CURL_CONSTANT(CURLINFO_FILETIME_T),
      GET_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_DOWNLOAD),
      GET_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_DOWNLOAD_T),
      GET_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_UPLOAD),
      GET_CURL_CONSTANT(CURLINFO_CONTENT_LENGTH_UPLOAD_T),
      GET_CURL_CONSTANT(CURLINFO_STARTTRANSFER_TIME),
      GET_CURL_CONSTANT(CURLINFO_CONTENT_TYPE),
      GET_CURL_CONSTANT(CURLINFO_REDIRECT_TIME),
      GET_CURL_CONSTANT(CURLINFO_REDIRECT_COUNT),
      GET_CURL_CONSTANT(CURLINFO_HTTP_CONNECTCODE),
      GET_CURL_CONSTANT(CURLINFO_HTTPAUTH_AVAIL),
      GET_CURL_CONSTANT(CURLINFO_PROXYAUTH_AVAIL),
      GET_CURL_CONSTANT(CURLINFO_OS_ERRNO),
      GET_CURL_CONSTANT(CURLINFO_NUM_CONNECTS),
      GET_CURL_CONSTANT(CURLINFO_COOKIELIST),
      GET_CURL_CONSTANT(CURLINFO_FTP_ENTRY_PATH),
      GET_CURL_CONSTANT(CURLINFO_REDIRECT_URL),
      GET_CURL_CONSTANT(CURLINFO_PRIMARY_IP),
      GET_CURL_CONSTANT(CURLINFO_APPCONNECT_TIME),
      GET_CURL_CONSTANT(CURLINFO_CERTINFO),
      GET_CURL_CONSTANT(CURLINFO_CONDITION_UNMET),
      GET_CURL_CONSTANT(CURLINFO_RTSP_SESSION_ID),
      GET_CURL_CONSTANT(CURLINFO_RTSP_CLIENT_CSEQ),
      GET_CURL_CONSTANT(CURLINFO_RTSP_SERVER_CSEQ),
      GET_CURL_CONSTANT(CURLINFO_RTSP_CSEQ_RECV),
      GET_CURL_CONSTANT(CURLINFO_PRIMARY_PORT),
      GET_CURL_CONSTANT(CURLINFO_LOCAL_IP),
      GET_CURL_CONSTANT(CURLINFO_LOCAL_PORT),
      GET_CURL_CONSTANT(CURLINFO_HTTP_VERSION),
      GET_CURL_CONSTANT(CURLINFO_PROXY_SSL_VERIFYRESULT),
      GET_CURL_CONSTANT(CURLINFO_PROTOCOL),
      GET_CURL_CONSTANT(CURLINFO_SCHEME),
      GET_CURL_CONSTANT(CURLINFO_TOTAL_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_NAMELOOKUP_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_CONNECT_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_PRETRANSFER_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_STARTTRANSFER_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_REDIRECT_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_APPCONNECT_TIME_T),
      GET_CURL_CONSTANT(CURLINFO_RETRY_AFTER),

      // CURL FILE TYPES
      GET_CURL_CONSTANT(CURLFILETYPE_FILE),
      GET_CURL_CONSTANT(CURLFILETYPE_DIRECTORY),
      GET_CURL_CONSTANT(CURLFILETYPE_SYMLINK),
      GET_CURL_CONSTANT(CURLFILETYPE_DEVICE_CHAR),
      GET_CURL_CONSTANT(CURLFILETYPE_NAMEDPIPE),
      GET_CURL_CONSTANT(CURLFILETYPE_SOCKET),
      GET_CURL_CONSTANT(CURLFILETYPE_DOOR),

      // CURL AUTH VALUES
      GET_CURL_CONSTANT(CURLAUTH_NONE),
      GET_CURL_CONSTANT(CURLAUTH_BASIC),
      GET_CURL_CONSTANT(CURLAUTH_DIGEST),
      GET_CURL_CONSTANT(CURLAUTH_NEGOTIATE),
      GET_CURL_CONSTANT(CURLAUTH_GSSNEGOTIATE),
      GET_CURL_CONSTANT(CURLAUTH_NTLM),
      GET_CURL_CONSTANT(CURLAUTH_DIGEST_IE),
      GET_CURL_CONSTANT(CURLAUTH_NTLM_WB),
      GET_CURL_CONSTANT(CURLAUTH_BEARER),
      GET_CURL_CONSTANT(CURLAUTH_ONLY),
      GET_CURL_CONSTANT(CURLAUTH_ANY),
      GET_CURL_CONSTANT(CURLAUTH_ANYSAFE),
      //ssh
      GET_CURL_CONSTANT(CURLSSH_AUTH_ANY),
      GET_CURL_CONSTANT(CURLSSH_AUTH_NONE),
      GET_CURL_CONSTANT(CURLSSH_AUTH_PUBLICKEY),
      GET_CURL_CONSTANT(CURLSSH_AUTH_PASSWORD),
      GET_CURL_CONSTANT(CURLSSH_AUTH_HOST),
      GET_CURL_CONSTANT(CURLSSH_AUTH_KEYBOARD),
      GET_CURL_CONSTANT(CURLSSH_AUTH_AGENT),
      GET_CURL_CONSTANT(CURLSSH_AUTH_GSSAPI),
      GET_CURL_CONSTANT(CURLSSH_AUTH_DEFAULT),
      // usessl
      GET_CURL_CONSTANT(CURLUSESSL_NONE),
      GET_CURL_CONSTANT(CURLUSESSL_TRY),
      GET_CURL_CONSTANT(CURLUSESSL_CONTROL),
      GET_CURL_CONSTANT(CURLUSESSL_ALL),

      {"version",    false, __curl_version},

      {NULL,    false, NULL},
  };

  static b_func_reg module_functions[] = {
      {"easy_init",   true,  GET_MODULE_METHOD(curl__easy_init)},
      {"easy_cleanup",   true,  GET_MODULE_METHOD(curl__easy_cleanup)},
      {"easy_reset",   true,  GET_MODULE_METHOD(curl__easy_reset)},
      {"easy_setopt",   true,  GET_MODULE_METHOD(curl__easy_setopt)},
      {"easy_perform",   true,  GET_MODULE_METHOD(curl__easy_perform)},
      {"easy_getinfo",   true,  GET_MODULE_METHOD(curl__easy_getinfo)},
      {"easy_escape",   true,  GET_MODULE_METHOD(curl__easy_escape)},
      {"easy_unescape",   true,  GET_MODULE_METHOD(curl__easy_unescape)},
      {"getdate",   true,  GET_MODULE_METHOD(curl__getdate)},
      {"mime_init",   true,  GET_MODULE_METHOD(curl__mime_init)},
      {"mime_addpart",   true,  GET_MODULE_METHOD(curl__mime_addpart)},
      {"mime_name",   true,  GET_MODULE_METHOD(curl__mime_name)},
      {"mime_data",   true,  GET_MODULE_METHOD(curl__mime_data)},
      {"mime_type",   true,  GET_MODULE_METHOD(curl__mime_type)},
      {"mime_encoding",   true,  GET_MODULE_METHOD(curl__mime_encoding)},
      {"mime_filedata",   true,  GET_MODULE_METHOD(curl__mime_filedata)},
      {"mime_subparts",   true,  GET_MODULE_METHOD(curl__mime_subparts)},
      {"slist_create",   true,  GET_MODULE_METHOD(curl__slist_create)},
      {"slist_free",   true,  GET_MODULE_METHOD(curl__slist_free)},
      {NULL,    false, NULL},
  };

  static b_module_reg module = {
      .name = "_curl",
      .fields = module_fields,
      .functions = module_functions,
      .classes = NULL,
      .preloader = &b__curl_module_preloader,
      .unloader = &b__curl_module_unloader
  };

  return &module;
}
